use std::error::Error;
use std::fs::File;
use std::io::{self, Read, Write};
use std::os::unix::io::{AsRawFd, FromRawFd, RawFd};
use std::os::unix::net::UnixDatagram;
use std::path::PathBuf;
use std::process;
use std::rc::Rc;

use dbus;
use libc;
use libslirp;
#[cfg(feature = "libsystemd")]
use libsystemd::daemon::{self, NotifyState};
use mio::unix::EventedFd;
use mio::*;
use nix::sched::{setns, CloneFlags};
use structopt::{clap::ArgGroup, StructOpt};

#[macro_use]
extern crate lazy_static;

mod tun;

#[derive(Debug, StructOpt)]
#[structopt(
    name = "libslirp-helper",
    about = "slirp helper process",
    rename_all = "kebab-case",
    group = ArgGroup::with_name("verb").required(true)
)]
struct Opt {
    /// Activate debug mode
    #[structopt(long)]
    debug: bool,
    /// Print capabilities
    #[structopt(long, group = "verb")]
    print_capabilities: bool,
    /// Exit with parent process
    #[structopt(long)]
    exit_with_parent: bool,
    /// DBus bus address
    #[structopt(long)]
    dbus_address: Option<String>,
    /// Helper instance ID
    #[structopt(long, name = "id")]
    dbus_id: Option<String>,
    /// Incoming migration data from DBus
    #[structopt(long)]
    dbus_incoming: bool,
    /// Unix datagram socket path
    #[structopt(long, parse(from_os_str), group = "verb")]
    socket_path: Option<PathBuf>,
    /// Unix datagram socket file descriptor
    #[structopt(long, group = "verb")]
    fd: Option<i32>,
    /// Incoming migration data
    #[structopt(long)]
    incoming_fd: Option<i32>,
    /// Set DHCP NBP URL (ex: tftp://10.0.0.1/my-nbp)
    #[structopt(long, name = "url")]
    dhcp_nbp: Option<String>,

    /// Path to network namespace to join.
    #[structopt(long)]
    netns: Option<PathBuf>,
    /// Interface name, such as "tun0".
    #[structopt(long, group = "verb")]
    interface: Option<String>,

    #[structopt(flatten)]
    slirp: libslirp::Opt,
}

fn set_exit_with_parent() {
    #[cfg(any(target_os = "linux", target_os = "android"))]
    unsafe {
        libc::prctl(libc::PR_SET_PDEATHSIG, libc::SIGTERM, 0, 0, 0);
    }
}

const DBUS_TOKEN: Token = Token(10_000_000);

fn slirp_state_read<'a, R: Read>(
    slirp: &libslirp::MioHandler<'a>,
    reader: &mut R,
) -> Result<(), Box<dyn Error>> {
    let mut buf = [0; 4];
    reader.read(&mut buf)?;
    let in_version = i32::from_be_bytes(buf);
    if in_version > libslirp::state_version() {
        return Err(format!(
            "Incompatible migration data version: {} > {}",
            in_version,
            libslirp::state_version()
        )
        .into());
    }

    slirp.ctxt.state_read(in_version, reader)?;
    slirp.register();
    Ok(())
}

fn print_capabilities() -> Result<(), Box<dyn Error>> {
    io::stdout().write_all(
        r#"{
  "type": "slirp-helper",
  "features": [
    "dbus-address",
    "dhcp",
    "exit-with-parent",
    "migrate",
    "tftp",
    "ipv4",
    "ipv6",
    "netns",
    "notify-socket",
    "restrict"
  ]
}
"#
        .as_bytes(),
    )?;

    Ok(())
}

fn set_netns(fd: RawFd) -> Result<(), nix::Error> {
    setns(fd, CloneFlags::CLONE_NEWNET)
}

lazy_static! {
    // XXX: dbus.rs is boring
    static ref POLL: Poll = Poll::new().unwrap();
}

fn main() -> Result<(), Box<dyn Error>> {
    let m = Opt::clap().get_matches();
    let mut opt = Opt::from_clap(&m);
    if opt.debug {
        dbg!(&opt);
    }
    if opt.print_capabilities {
        return print_capabilities();
    }

    if m.occurrences_of("dhcp-start") == 0 {
        let dhcp_start = opt.slirp.ipv4.net.nth(15).expect("Invalid --net");
        opt.slirp.ipv4.dhcp_start = dhcp_start;
    }

    if let Some(url) = &opt.dhcp_nbp {
        let url = url::Url::parse(url)?;
        if url.scheme() != "tftp" {
            return Err("Invalid NBP URL".into());
        }
        opt.slirp.tftp.name = Some(url.host_str().unwrap().to_string());
        opt.slirp.tftp.bootfile = Some(url.path().to_string());
    }

    let mut main_netns = None;
    if let Some(netns) = &opt.netns {
        main_netns = Some(File::open("/proc/self/ns/net")?);
        let netns = File::open(netns)?;
        set_netns(netns.as_raw_fd())?;
        opt.interface.get_or_insert("tun0".to_string());
    }

    let stream = match &opt {
        Opt { fd: Some(fd), .. } => unsafe { UnixDatagram::from_raw_fd(*fd) },
        Opt {
            socket_path: Some(path),
            ..
        } => UnixDatagram::bind(path)?,
        Opt {
            interface: Some(tun),
            ..
        } => tun::open(tun)?,
        _ => return Err("Missing a socket argument".into()),
    };

    if let Some(netns) = main_netns {
        set_netns(netns.as_raw_fd())?;
    }

    if opt.exit_with_parent {
        set_exit_with_parent();
    }

    let slirp = Rc::new(libslirp::MioHandler::new(&opt.slirp, &POLL, stream));

    let mut dbus_fd = -1;
    let mut dbus = if let Some(dbus_addr) = opt.dbus_address {
        if opt.dbus_id.is_none() {
            return Err("You must specify an id with DBus".into());
        }

        let mut c = dbus::channel::Channel::open_private(&dbus_addr)?;
        if c.register().is_err() && opt.debug {
            eprintln!("Failed to register to the DBus bus, ignoring");
        }
        c.set_watch_enabled(true);
        dbus_fd = c.watch().fd;

        let c = dbus::blocking::LocalConnection::from(c);
        c.request_name(
            &format!("org.freedesktop.Slirp1_{}", process::id()),
            false,
            false,
            false,
        )?;
        c.request_name("org.qemu.VMState1", false, false, false)?;

        let f = dbus::tree::Factory::new_fn::<()>();
        // XXX: could use dbus-rs path get_data() ?
        let slirp = slirp.clone();
        let slirp1 = slirp.clone();
        let slirp2 = slirp.clone();
        let dbus_id = opt.dbus_id.unwrap().clone();
        let tree = f
            .tree(())
            .add(
                f.object_path("/org/freedesktop/Slirp1/Helper", ())
                    .introspectable()
                    .add(
                        f.interface("org.freedesktop.Slirp1.Helper", ()).add_m(
                            f.method("GetInfo", (), move |m| {
                                let info = slirp.ctxt.connection_info();
                                let mret = m.msg.method_return().append1(info);
                                Ok(vec![mret])
                            })
                            .outarg::<&str, _>("info"),
                        ),
                    ),
            )
            .add(
                f.object_path("/org/qemu/VMState1", ())
                    .introspectable()
                    .add(
                        f.interface("org.qemu.VMState1", ())
                            .add_p(
                                f.property::<&str, _>("Id", ())
                                    .emits_changed(dbus::tree::EmitsChangedSignal::False)
                                    .on_get(move |i, _m| {
                                        i.append(&dbus_id);
                                        Ok(())
                                    }),
                            )
                            .add_m(
                                f.method("Save", (), move |m| {
                                    let mut data = libslirp::state_version().to_be_bytes().to_vec();
                                    let mut state = match slirp1.ctxt.state_get() {
                                        Ok(s) => s,
                                        Err(e) => {
                                            return Err(dbus::tree::MethodErr::failed(
                                                &e.to_string(),
                                            ))
                                        }
                                    };
                                    data.append(&mut state);
                                    let mret = m.msg.method_return().append1(data);
                                    Ok(vec![mret])
                                })
                                .outarg::<Vec<u8>, _>("data"),
                            )
                            .add_m(
                                f.method("Load", (), move |m| {
                                    let data: Vec<u8> = m.msg.read1()?;
                                    if let Err(e) = slirp_state_read(&slirp2, &mut data.as_slice())
                                    {
                                        return Err(dbus::tree::MethodErr::failed(&e.to_string()));
                                    }
                                    let mret = m.msg.method_return();
                                    Ok(vec![mret])
                                })
                                .inarg::<Vec<u8>, _>("data"),
                            ),
                    ),
            )
            .add(f.object_path("/", ()).introspectable());
        tree.start_receive(&c);

        Some(c)
    } else {
        None
    };

    if dbus_fd != -1 {
        POLL.register(
            &EventedFd(&dbus_fd),
            DBUS_TOKEN,
            Ready::readable(),
            PollOpt::level(),
        )?;
    }

    if opt.dbus_incoming && opt.incoming_fd.is_some() {
        return Err("Invalid multiple incoming paths.".into());
    }

    let mut events = Events::with_capacity(1024);
    let mut duration = None;

    if let Some(fd) = opt.incoming_fd {
        let mut f = unsafe { File::from_raw_fd(fd) };
        slirp_state_read(&slirp, &mut f)?;
    } else if !opt.dbus_incoming {
        slirp.register();
    }

    #[cfg(feature = "libsystemd")]
    daemon::notify(true, &[NotifyState::Ready])?;

    loop {
        if opt.debug {
            dbg!(duration);
        }

        POLL.poll(&mut events, duration)?;
        duration = slirp.dispatch(&events)?;
        if let Some(ref mut dbus) = dbus {
            for event in &events {
                match event.token() {
                    DBUS_TOKEN => {
                        dbus.process(std::time::Duration::from_millis(0))?;
                    }
                    _ => {
                        continue;
                    }
                }
            }
        }
    }
}
